#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import time
import yaml

import upload_package
import watch_ppa

from ci_utils import (
    amqp_utils,
    data_store,
    dump_stack,
)

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

TIME_BETWEEN_CHECKS = 60


def get_auth_config():
    DEFAULT_CONFIG_YAML = os.path.join(
        os.path.dirname(__file__), '../../unit_config')
    with open(DEFAULT_CONFIG_YAML) as f:
        config = yaml.safe_load(f)
    return config


def create_result_file(datastore, content):
    '''Creates a result file based on the final upload and build status.

    The input content has the following format:
        {
            'message': 'The high level status of the upload',
            'subticket_status': [{
                'name': source_package_name,
                'version': source_package_version,
                'step': numeric_workflow_step,
                'step_text': text_workflow_step,
                'status': numeric_workflow_status,
                'status_text': text_workflow_status,
                'message': subticket_specific_message,
            }]
        }

    The return is an artifact entry for the progress message.'''
    package_build_result_filename = 'package_build.output.log'
    body = '{}\n\n'.format(content['message'])
    body += 'Uploaded source packages:\n'
    for subticket_status in content['subticket_status']:
        body += 'name: {}\n'.format(subticket_status['name'])
        body += 'version: {}\n'.format(subticket_status['version'])
        body += 'status: {}\n'.format(subticket_status['status_text'])
        if subticket_status.get('message'):
            body += 'message: {}\n'.format(subticket_status['message'])
    location = datastore.put_file(
        package_build_result_filename, str(body), 'text/plain')
    return {
        'name': package_build_result_filename,
        'type': 'LOGS',
        'reference': location,
    }


def on_message(msg):
    log.info('on_message: {}'.format(msg.body))
    params = json.loads(msg.body)
    ticket_id = params['ticket_id']
    subtickets = params['subtickets']
    series = params['series']
    ppa = params['ppa']
    archive = params['archive']
    trigger = params['progress_trigger']

    # Setup the output data to send back to the caller
    out_data = {
        'message': 'Build in progress.',
        'ppa': ppa,
        'archive': archive,
        'artifacts': [],
    }
    amqp_utils.progress_update(trigger, params)

    try:
        datastore = data_store.create_for_ticket(ticket_id, get_auth_config())

        upload_list = upload_package.upload_source_packages(ppa, subtickets)
        log.info('upload_list: {}'.format(upload_list))

        # Check that an upload isn't going to fail due to version
        (failed, subticket_status) = watch_ppa.check_ppa(
            series, ppa, archive, None, upload_list)
        if failed:
            # This is the end of the line, the upload is blocked.
            out_data['subticket_status'] = subticket_status
            message = 'Source package upload failed.'
            log.error(message)
            out_data['message'] = message
            out_data['artifacts'].append(
                create_result_file(datastore, out_data))
            amqp_utils.progress_failed(trigger, out_data)
        else:
            # Set ret to 1, the upload is failed until it succeeds
            ret = 1
            start_time = time.time()
            while True:
                (ret, subticket_status) = watch_ppa.watch_ppa(
                    start_time, series, ppa, archive, None, upload_list)
                log.info('subticket_status: {}'.format(subticket_status))
                out_data['subticket_status'] = subticket_status
                if ret == -1:
                    # -1 indicates that the task is not finished
                    # Send regular progress back to the lander
                    amqp_utils.progress_update(trigger, out_data)
                    log.info('Going to sleep for '
                             '{}'.format(TIME_BETWEEN_CHECKS))
                    time.sleep(TIME_BETWEEN_CHECKS)
                else:
                    log.info('All done')
                    break

            if ret == 1:
                # '1' indicates a failure occurred during the upload/build
                message = 'A package failed to upload or build.'
                log.error(message)
                out_data['message'] = message
                out_data['artifacts'].append(
                    create_result_file(datastore, out_data))
                amqp_utils.progress_failed(trigger, out_data)
            else:
                # '0' indicates the build was successful
                message = 'The package build was successful.'
                log.error(message)
                out_data['message'] = message
                out_data['artifacts'].append(
                    create_result_file(datastore, out_data))
                amqp_utils.progress_completed(trigger, out_data)
    except (KeyboardInterrupt, Exception) as e:
        message = 'Exception: {}'.format(e)
        log.error(message)
        out_data['message'] = message
        amqp_utils.progress_failed(trigger, out_data)
        if isinstance(e, KeyboardInterrupt):
            raise  # re-raise so amqp_utils.process_queue can exit
    finally:
        # remove from queue so request becomes completed
        log.info('Acking the request: {}'.format(msg.body))
        msg.channel.basic_ack(msg.delivery_tag)


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    config = amqp_utils.get_config()
    if not config:
        exit(1)  # the get_config code prints an error
    amqp_utils.process_queue(config, 'bsbuilder', on_message)
    log.info('exiting')
