#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import sys
from cStringIO import StringIO


logging.basicConfig(level=logging.INFO)
log = logging.getLogger('Imagebuilder')
logstringio = StringIO()
logstream = logging.StreamHandler(logstringio)
formatter = logging.Formatter(
    '[%(asctime)s] %(name)s:%(levelname)s:%(message)s')
logstream.setFormatter(formatter)
logstream.setLevel(logging.INFO)
log.addHandler(logstream)

from ci_utils import amqp_utils
from ci_utils import dump_stack
from ci_utils import data_store

from imagebuilder.cloud_image import modify_image
from imagebuilder.cloud_image import get_auth_config


def on_message(msg):
    log.info('on_message: {0}'.format(msg.body))
    params = json.loads(msg.body)
    image = params['image']
    repos = params['ppa_list']
    series = params['series']
    packages = params['package_list']
    trigger = params['progress_trigger']
    ticket = params['ticket_id']
    try:
        exc_type = None
        store = data_store.create_for_ticket(ticket, get_auth_config())
        amqp_utils.progress_update(trigger, params)

        def status_cb(msg):
            log.info(msg)
            amqp_utils.progress_update(trigger, {'message': msg})

        image_id = modify_image(image, repos, series, packages, status_cb)

    except (KeyboardInterrupt, Exception) as e:
        exc_type, val, tb = sys.exc_info()
        log.exception('Image build failed:')
        if isinstance(e, KeyboardInterrupt):
            raise  # re-raise so amqp_utils.process_queue can exit

    finally:
        # remove from queue so request becomes completed
        msg.channel.basic_ack(msg.delivery_tag)
        if exc_type:
            amqp_status_func = amqp_utils.progress_failed
            return_status = {
                'message': '%s: %s' % (exc_type.__name__, val),
            }
        else:
            amqp_status_func = amqp_utils.progress_completed
            return_status = {
                'image_id': image_id,
            }

        try:
            ds_exc_type = None
            log_name = 'image_builder.output.log'
            logurl = store.put_file(log_name,
                                    logstringio.getvalue(),
                                    content_type='text/plain')
            artifacts = [{
                'name': log_name,
                'reference': '%s' % logurl,
                'type': 'LOGS',
            }]
            return_status.update({'artifacts': artifacts})
        except (KeyboardInterrupt, Exception) as e:
            ds_exc_type, ds_val, ds_tb = sys.exc_info()
            log.exception('data_store put_file failed:')
            if isinstance(e, KeyboardInterrupt):
                raise  # re-raise so amqp_utils.process_queue can exit
        finally:
            if ds_exc_type:
                return_status.update(
                    {'error': '%s: %s' % (ds_exc_type.__name__, ds_val)})
            amqp_status_func(trigger, return_status)


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    config = amqp_utils.get_config()
    if not config:
        exit(1)  # the get_config code prints an error
    amqp_utils.process_queue(config, 'imagebuilder', on_message)
    log.info('exiting')
