#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging

from ci_utils import amqp_utils
from ci_utils import dump_stack
from ci_utils import data_store
import tstrun
from tstrun import testbed

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def on_message(msg):
    log.info('on_message: %s', msg.body)
    params = json.loads(msg.body)
    ticket = params['ticket_id']
    progress_queue = params['progress_trigger']
    image_id = params['image_id']
    package_list = params['package_list']
    amqp_utils.progress_update(progress_queue, params)

    conf = tstrun.get_auth_config()

    def status_cb(msg):
        amqp_utils.progress_update(progress_queue, {'message': msg})

    # The tests will succeed unless they fail ;)
    notify = amqp_utils.progress_completed
    results = {
        'artifacts': [],
    }
    try:
        test_bed = None
        store = data_store.create_for_ticket(ticket, conf)
        flavors = conf['tr_flavors']
        test_bed = testbed.TestBed('testbed-{}'.format(progress_queue),
                                   flavors, image_id, status_cb)
        test_bed.setup()

        # FIXME: There is a potentially confusing failure mode here: by reusing
        # the same testbed for all packages, we may miss bad dependencies for
        # one package because that missing dependency is provided by another
        # package. It's vague and unclear that we should care about this edge
        # case at this point -- vila 2014-02-01
        for package in package_list:
            status_cb('testing %s' % package)
            # uci-vms shell adt-run ...  --- adt-virt-null
            return_code, subunit_stream = test_bed.run_test(package)
            # uci-vms shell collect test results
            artifact_name = 'test_runner.{}-subunit-stream'.format(package)
            subunit_url = store.put_file(artifact_name,
                                         subunit_stream)
            artifact = {
                'name': artifact_name,
                'reference': subunit_url,
                'type': 'RESULTS',
            }
            results['artifacts'].append(artifact)
            # 0 is success, 8 is skipped and considered a success
            if return_code not in (0, 8):
                # At least one test failed
                notify = amqp_utils.progress_failed
    except (KeyboardInterrupt, Exception) as e:
        results['message'] = str(e)
        notify = amqp_utils.progress_failed
        if isinstance(e, KeyboardInterrupt):
            raise  # re-raise so amqp_utils.process_queue can exit
    finally:
        # All tests have been run, report success or failure
        notify(progress_queue, results)
        # Request completed, remove from queue

        # FIXME: There is still a hole somewhere, the ack below won't happen if
        # an exception occurs while calling 'notify' just above, tricky testing
        # ahead... -- vila 2014-02-06
        msg.channel.basic_ack(msg.delivery_tag)
        if test_bed:
            test_bed.teardown()


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    config = amqp_utils.get_config()
    if not config:
        exit(1)  # the get_config code prints an error
    amqp_utils.process_queue(config, tstrun.internal_rabbit_queue, on_message)
    log.info('exiting')
